#include "testlib.h"
#include <unordered_set>

struct Edge {
    int32_t u, v, t;
};

struct Test {
    int32_t n;
    std::vector<Edge> edges;
};

std::ostream& operator<<(std::ostream& out, const Test& test) {
    out << test.n << ' ' << test.edges.size() << '\n';
    for (const Edge& edge : test.edges)
        out << edge.u + 1 << ' ' << edge.v + 1 << ' ' << edge.t << '\n';
    return out;
}

Test generate_random_test() {
    Test test;
    const int32_t MIN_N = opt<int32_t>("min_n");
    const int32_t MAX_N = opt<int32_t>("max_n");
    test.n = rnd.next(MIN_N, MAX_N);
    const int32_t MIN_M = opt<int32_t>("min_m");
    const int32_t MAX_M = opt<int32_t>("max_m");
    test.edges.resize(rnd.next(MIN_M, MAX_M));
    const int32_t MIN_T = opt<int32_t>("min_t");
    const int32_t MAX_T = opt<int32_t>("max_t");
    for (Edge& edge : test.edges) {
        edge.u = rnd.next(0, test.n - 1);
        edge.v = rnd.next(0, test.n - 1);
        edge.t = rnd.next(MIN_T, MAX_T);
    }
    return test;
}

Test generate_almost_max_ans_test() {
    Test test;
    const int32_t MIN_N = opt<int32_t>("min_n");
    const int32_t MAX_N = opt<int32_t>("max_n");
    const int32_t MIN_M = opt<int32_t>("min_m");
    const int32_t MAX_M = opt<int32_t>("max_m");
    const int32_t MIN_T = opt<int32_t>("min_t");
    const int32_t MAX_T = opt<int32_t>("max_t");
    const int32_t REAL_MIN_N = std::max(2, (MIN_M + 5) / 2);
    const int32_t REAL_MAX_N = std::min(MAX_N, (MAX_M + 4) / 2);
    const int32_t real_n = rnd.next(REAL_MIN_N, REAL_MAX_N);
    test.n = rnd.next(std::max(MIN_N, real_n), MAX_N);
    std::vector<int32_t> v(test.n);
    for (int32_t i = 0; i < test.n; ++i)
        v[i] = i;
    shuffle(v.begin() + 1, v.end() - 1);
    std::swap(v[real_n - 1], v.back());
    std::swap(v[real_n - 2], v.front());
    const int32_t m = real_n * 2 - 4;
    test.edges.resize(m);
    for (int32_t i = 1; i < real_n - 1; ++i) {
        test.edges[2 * i - 2] = {
                .u = v[i],
                .v = v[i - 1],
                .t = rnd.next(MIN_T, MAX_T),
        };
        test.edges[2 * i - 1] = {
                .u = v[i],
                .v = v[i + 1],
                .t = rnd.next(MIN_T, MAX_T),
        };
    }
    shuffle(test.edges.begin(), test.edges.end());
    return test;
}

Test generate_dense_dag_test() {
    Test test;
    const int32_t MIN_N = opt<int32_t>("min_n");
    const int32_t MAX_N = opt<int32_t>("max_n");
    test.n = rnd.next(MIN_N, MAX_N);
    const int32_t MIN_M = opt<int32_t>("min_m");
    const int32_t MAX_M = opt<int32_t>("max_m");
    test.edges.resize(rnd.next(MIN_M, MAX_M));
    const int32_t MIN_D = opt<int32_t>("min_d");
    const int32_t MAX_D = std::min<int32_t>(opt<int32_t>("max_d"), ssize(test.edges) + 1);
    const int32_t d = rnd.next(MIN_D, MAX_D);
    const int64_t MAX_P = opt<int64_t>("max_p");
    const int32_t MAX_T = opt<int32_t>("max_t");
    std::vector<int64_t> potentials(d);
    while (true) {
        for (int32_t i = 1; i < d; ++i)
            potentials[i] = potentials[i - 1] + rnd.next(1, MAX_T);
        for (int32_t i = 1; i < d; ++i)
            if (potentials[i] > MAX_P)
                potentials[i] = rnd.next(static_cast<int64_t>(1), MAX_P);
        std::sort(potentials.begin() + 1, potentials.end());
        bool small_jump = true;
        for (int32_t i = 0; i + 1 < d; ++i)
            if (potentials[i + 1] > potentials[i] + MAX_T) {
                small_jump = false;
                break;
            }
        if (small_jump)
            break;
    }
    std::vector<std::pair<int32_t, int32_t>> admissible_endpoints(d);
    std::vector<int64_t> prefix_endpoints(d);
    for (int32_t i = 1; i < d; ++i) {
        admissible_endpoints[i] = admissible_endpoints[i - 1];
        while (potentials[i] > potentials[admissible_endpoints[i].first] + MAX_T)
            ++admissible_endpoints[i].first;
        while (potentials[i] > potentials[admissible_endpoints[i].second])
            ++admissible_endpoints[i].second;
        prefix_endpoints[i] = prefix_endpoints[i - 1] + admissible_endpoints[i].second - admissible_endpoints[i].first;

        int32_t v = rnd.next(admissible_endpoints[i].first, admissible_endpoints[i].second - 1);
        test.edges[i - 1] = {
                .u = i,
                .v = v,
                .t = static_cast<int32_t>(potentials[i] - potentials[v]),
        };
    }
    const auto dag_edges_part = opt<double>("dag_edges_part");
    const int32_t dag_edges = (ssize(test.edges) - (d - 1)) * dag_edges_part;
    ensuref(dag_edges >= 0, "%s", std::format("dag_edges == {} < 0", dag_edges).c_str());
    for (int32_t i = d - 1; i < d - 1 + dag_edges; ++i) {
        // choose an index of an edge
        int64_t edge_index = rnd.next(static_cast<int64_t>(0), prefix_endpoints.back() - 1);
        test.edges[i].u = upper_bound(prefix_endpoints.begin(), prefix_endpoints.end(), edge_index) - prefix_endpoints.begin();
        test.edges[i].v = rnd.next(admissible_endpoints[test.edges[i].u].first, admissible_endpoints[test.edges[i].u].second - 1);
        test.edges[i].t = potentials[test.edges[i].u] - potentials[test.edges[i].v];
    }

    const auto killer_vertices_part = opt<double>("killer_vertices_part");
    const int32_t n_killer_vertices = (d - 1) * killer_vertices_part;
    const auto killer_edges_part = opt<double>("killer_edges_part");
    const int32_t n_killer_edges = (test.edges.size() - (d - 1 + dag_edges)) * killer_edges_part;
    ensuref(n_killer_edges == 0 || n_killer_vertices > 0, "%s", std::format("n_killer_vertices == {} but n_killer_edges == {}", n_killer_vertices, n_killer_edges).c_str());
    std::vector<int32_t> killer_vertices;
    std::unordered_set<int32_t> killer_vertices_unordered_set;
    while (killer_vertices_unordered_set.size() < n_killer_vertices) {
        int32_t killer_vertex = rnd.next(0, d - 1);
        killer_vertices_unordered_set.insert(killer_vertex);
    }
    killer_vertices = std::vector(killer_vertices_unordered_set.begin(), killer_vertices_unordered_set.end());
    for (int32_t i = d - 1 + dag_edges; i < d - 1 + dag_edges + n_killer_edges; ++i) {
        test.edges[i].u = rnd.any(killer_vertices.begin(), killer_vertices.end());
        test.edges[i].v = rnd.next(d, test.n - 1);
        test.edges[i].t = rnd.next(1, MAX_T);
    }
    for (int32_t i = d - 1 + dag_edges + n_killer_edges; i < test.edges.size(); ++i)
        while (true) {
            test.edges[i].u = rnd.next(0, test.n - 1);
            test.edges[i].v = rnd.next(0, test.n - 1);
            int32_t min_t = 1;
            if (test.edges[i].u < d && test.edges[i].v < d) {
                int64_t t = potentials[test.edges[i].u] - potentials[test.edges[i].v];
                if (t >= MAX_T)
                    continue;
                min_t = std::max<int64_t>(min_t, t + 1);
            }
            if (test.edges[i].u >= d && test.edges[i].v < d)
                continue;
            if (test.edges[i].u < d && test.edges[i].v >= d)
                continue;
            test.edges[i].t = rnd.next(min_t, MAX_T);
            break;
        }
    shuffle(test.edges.begin(), test.edges.end());
    std::vector<int32_t> permutation(test.n);
    for (int32_t i = 0; i < test.n; ++i)
        permutation[i] = i;
    std::swap(permutation[0], permutation[test.n - 2]);
    shuffle(permutation.begin(), permutation.end() - 2);
    int32_t zero_index = rnd.next(1, d - 1);
    std::swap(permutation[zero_index], permutation[test.n - 2]);
    std::swap(permutation[0], permutation[test.n - 1]);
    for (Edge& e : test.edges) {
        e.u = permutation[e.u];
        e.v = permutation[e.v];
    }
    return test;
}

Test generate_multilayer_test() {
    Test test;
    const int32_t MIN_N = opt<int32_t>("min_n");
    const int32_t MAX_N = opt<int32_t>("max_n");
    test.n = rnd.next(MIN_N, MAX_N);
    const int32_t MIN_M = opt<int32_t>("min_m");
    const int32_t MAX_M = opt<int32_t>("max_m");
    const int32_t m = rnd.next(MIN_M, MAX_M);
    test.edges.reserve(m);
    const int32_t MIN_L = opt<int32_t>("min_l");
    const int32_t MAX_L = opt<int32_t>("max_l");
    const int32_t MIN_K = opt<int32_t>("min_k");
    const int32_t MAX_K = opt<int32_t>("max_k");
    const int32_t k = rnd.next(MIN_K, MAX_K);
    const int32_t l = rnd.next(MIN_L, MAX_L);
    const int32_t MAX_T = opt<int32_t>("max_t");
    const auto dead_part = opt<double>("dead_part");
    const auto ill_part = opt<double>("ill_part");
    const auto start_part = opt<double>("start_part");
    const auto end_part = opt<double>("end_part");

    const int32_t dead_vertices = std::min<int32_t>(ceil(dead_part * (test.n - 1)), test.n - l);
    const int32_t alive_vertices = test.n - dead_vertices;
    //const int32_t l = rnd.next(MIN_L, std::min(MAX_L, alive_vertices));

    const int32_t ill_layer = round(ill_part * (l - 1));
    const int32_t end_layer = round(end_part * (l - 1));
    const int32_t start_layer = std::max<int32_t>(round(start_part * (l - 1)), end_layer + 1);

    std::vector<int64_t> potential_bounds(l + 1);
    std::vector<int64_t> potential(alive_vertices);
    for (int32_t i = 0; i <= l; ++i)
        potential_bounds[i] = static_cast<int64_t>(MAX_T / (k + 1)) * i;
    std::vector<int32_t> layer_by_vertex(alive_vertices);
    std::vector<std::vector<int32_t>> layers(l);
    for (int32_t i = 0; i < alive_vertices; ++i) {
        if (i < l)
            layer_by_vertex[i] = i;
        else
            layer_by_vertex[i] = rnd.next(0, l - 1);
        layers[layer_by_vertex[i]].push_back(i);
        potential[i] = rnd.next(potential_bounds[layer_by_vertex[i]], potential_bounds[layer_by_vertex[i] + 1] - 1);
    }
    for (int32_t layer = 1; layer < l; ++layer)
        for (int32_t u : layers[layer]) {
            int32_t L = std::max(0, layer - k), R = layer - 1;
            if (L <= end_layer && R >= end_layer)
                L = end_layer;
            int32_t another_layer = rnd.next(L, R);
            int32_t v = another_layer == end_layer ?  layers[another_layer][0] : rnd.any(layers[another_layer]);
            test.edges.push_back({.u = u, .v = v, .t = static_cast<int32_t>(potential[u] - potential[v])});
        }
    for (int32_t i = std::ssize(test.edges); i < m; ++i) {
        int32_t edge_type = rnd.next(0, 2);
        if (edge_type == 0) {
            int32_t layer = rnd.next(1, l - 1);
            int32_t u = rnd.any(layers[layer]);
            int32_t another_layer = rnd.next(std::max(0, layer - k), layer - 1);
            int32_t v = rnd.any(layers[another_layer]);
            test.edges.push_back({.u = u, .v = v, .t = static_cast<int32_t>(potential[u] - potential[v])});
        } else if (edge_type == 1) {
            int32_t layer = rnd.next(0, l - 2);
            int32_t u = rnd.any(layers[layer]);
            int32_t another_layer = rnd.next(layer + 1, std::min(l - 1, layer + k));
            int32_t v = rnd.any(layers[another_layer]);
            int32_t t = rnd.next(1, MAX_T);
            test.edges.push_back({.u = u, .v = v, .t = t});
        } else {
            int32_t u = rnd.any(layers[ill_layer]);
            int32_t v = rnd.next(alive_vertices, test.n - 1);
            int32_t t = rnd.next(1, MAX_T);
            test.edges.push_back({.u = u, .v = v, .t = t});
        }
    }
    shuffle(test.edges.begin(), test.edges.end());
    std::vector<int32_t> permutation(test.n);
    for (int32_t i = 0; i < test.n; ++i)
        permutation[i] = i;
    std::swap(permutation[0], permutation[test.n - 2]);
    shuffle(permutation.begin(), permutation.end() - 2);
    std::swap(permutation[layers[end_layer][0]], permutation[test.n - 1]);
    std::swap(permutation[rnd.any(layers[start_layer])], permutation[test.n - 2]);
    for (Edge& e : test.edges) {
        e.u = permutation[e.u];
        e.v = permutation[e.v];
    }
    return test;
}

int main(int argc, char* argv[]) {
    registerGen(argc, argv, 1);
    const auto type = opt<std::string>("type");
    Test test;
    if (type == "random")
        test = generate_random_test();
    else if (type == "almost_max_ans")
        test = generate_almost_max_ans_test();
    else if (type == "dense_dag")
        test = generate_dense_dag_test();
    else if (type == "multilayer")
        test = generate_multilayer_test();
    else
        quitf(_fail, "%s", std::format("unknown test type: '{}'", type).c_str());
    std::cout << test;
}
